import logging
import requests
import getpass
import json
import re
from requests import HTTPError

from redash_tools.core.entities import Query, Dashboard

logger = logging.getLogger(__name__)


class RedashSession:
   
    def __init__(self, url=None, api_key=None):
        """
        initializes RedashSession
        url is Redash's url, api_key is the user's API key
        """
        if url is None:
            url = input('Введите url (включая http/https): ').strip('/\\ ')
        if api_key is None:
            api_key = getpass.getpass(f'Введите API-ключ со страницы {url}/users/me : ')
        self.url = url
        self.s = requests.Session()
        self.s.headers.update({'Authorization': f'Key {api_key}',
                               'Content-Type': 'application/json'})

    def make_url(self, *args):
        args = list(args)
        args.insert(0, self.url)
        return '/'.join(str(arg) for arg in args)

    def make_api_url(self, uri, *args):
        return self.make_url('api', uri, *args)
                
    #######################
    # get-methods section #  
    #######################
        
    def get(self, uri: str):
        """
        gets entity of given entity_type with given entity_id
        returns dict
        can raise JSONDecodeError, HTTPError
        """ 
        response = self.s.get(self.make_api_url(uri))
        response.raise_for_status()
        return response.json()
    
    def get_all(self, uri: str):
        """
        gets all entities for given entity_type
        queries without visualizations, dashboards without widgets
        returns list of dicts
        can raise JSONDecodeError, HTTPError, TypeError
        """
        response = self.s.get(self.make_api_url(uri))
        response.raise_for_status()
        response = response.json()
        if type(response) == dict:
            entities = []
            page_count = int(response['count'] / response['page_size']) + 2
            for page in range(1, page_count):
                response = self.s.get(self.make_api_url(uri), params={'page': page}).json()
                entities.extend(response['results'])
        elif type(response) == list:
            entities = response
        else:
            raise TypeError('Response must be either dict or list')
        return entities
    
    def get_data_sources(self, include_view_only=True):
        """
        gets dictionary of Data Sources
        not used in other methods, just for info purpose
        """
        data_sources = self.get_all('data_sources')
        ds_dict = {ds.pop('name'): {key: ds[key] for key in ('id', 'type', 'view_only')}
                   for ds in data_sources if include_view_only or not ds['view_only']}
        return ds_dict 
    
    def get_query(self, query_id):
        """
        gets query with given id
        returns object of class Query
        """
        q = self.get(f'queries/{query_id}')
        return Query.from_dict(q)
    
    def get_queries(self, query_ids):
        """
        gets queries with given ids
        returns list of Query objects
        """
        queries = []
        for query_id in query_ids:
            queries.append(self.get_query(query_id))
        return queries
    
    def get_dashboard(self, slug):
        """
        gets dashboard with given slug
        returns object of class Dashboard
        """
        d = self.get(f'dashboards/{slug}')
        return Dashboard.from_dict(d)

    def get_dashboards(self, slugs):
        """
        gets dashboards with given slugs
        returns list of Dashboard objects
        """
        dashboards = []
        for slug in slugs:
            dashboards.append(self.get_dashboard(slug))
        return dashboards

    ########################
    # find-methods section #
    ########################


    def find_by_conditions(self, uri: str, conditions: dict, regex=False, return_slugs=False):
        """
        finds ids of uri, matching all given conditions (default exact matching, set regexp=True for re.search)
        """
        entities = self.get_all(uri)
        return _find_by_conditions(entities, conditions, regex, return_slugs)


    ########################
    # post-methods section #
    ########################
    
    def post(self, uri: str, data: dict):
        """
        posts some dict data to entity_type
        """
        res = self.s.post(self.make_api_url(uri), data=json.dumps(data))
        res.raise_for_status()
        return res.json()
    
    ##########################
    # delete-methods section #
    ##########################
             
    def delete(self, uri: str, data=None):
        if data is None:
            res = self.s.delete(self.make_api_url(uri))
        else:
            res = self.s.delete(self.make_api_url(uri), data=json.dumps(data))
        res.raise_for_status()
        return res.json()

    # def delete_entity(self, entity: RedashEntity):
    #     self.delete(entity.make_uri())
    #
    # def delete_dashboard(self, slug: str):
    #     d = self.get('dashboards', slug)
    #     if d['can_edit']:
    #         ans = input(f'Вы действительно хотите удалить дашборд {self.make_url("dashboard", slug)}? '
    #                     'Это действие невозможно отменить. [y/n]')
    #         if ans.lower() == 'y':
    #             self.delete('dashboards', slug)
    #         else:
    #             logger.error('Операция прервана')
    #             return
    #     else:
    #         logger.error(f'Нет прав на удаление дашборда {self.make_url("dashboard", slug)}')
    #         return
        
    ##########################
    # change-methods section #
    ##########################

    def _change_access(self, entity_type: str, entity_ids: list, user_ids: list, grant: bool):
        error_ids = set()
        for entity_id in entity_ids:
            try:
                for user_id in user_ids:
                    uri = f'{entity_type}/{entity_id}/acl'
                    data = {'user_id': user_id, 'access_type': 'modify'}
                    self.post(uri, data) if grant else self.delete(uri, data)
            except HTTPError:
                error_ids.add(entity_id)
        if len(error_ids) > 0:
            raise UserWarning(f'Не удалось изменить права для {entity_type} {error_ids}')

    def grant_access(self, entity_type: str, entity_ids: list, user_ids: list):
        self._change_access(entity_type, entity_ids, user_ids, grant=True)

    def limit_access(self, entity_type: str, entity_ids: list, user_ids: list):
        self._change_access(entity_type, entity_ids, user_ids, grant=False)

    def _change_entities(self, entity_type: str, entity_ids: list, data):
        error_ids = set()
        for entity_id in entity_ids:
            try:
                self.post(f'{entity_type}/{entity_id}', data)
            except HTTPError:
                error_ids.add(entity_id)
        if len(error_ids) > 0:
            raise UserWarning(f'Не удалось изменить {entity_type} {error_ids}')

    def archive_queries(self, query_ids: list):
        self._change_entities('queries', query_ids, {'is_archived': True})

    def restore_queries(self, query_ids: list):
        self._change_entities('queries', query_ids, {'is_archived': False})

    def tag_queries(self, query_ids: list, tags: list):
        self._change_entities('queries', query_ids, {'tags': tags})

    def schedule_queries(self, query_ids: list, interval_sec: int):
        schedule = {'interval': interval_sec,
                         'until': None,
                         'day_of_week': None,
                         'time': None}
        self._change_entities('queries', query_ids, {'schedule': schedule})
        
    def replace_query_sql(self, query_ids: list, str_from, str_to, regex=False):
        queries = self.get_queries(query_ids)
        queries = [q.replace_sql(str_from, str_to, regex) for q in queries]
        for q in queries:
            self._change_entities('queries', [q.id], {'query': q.query})


def _test_connection(url, s):
    r = s.get(f'{url}/api/queries')
    r.raise_for_status()
    r.json()


def _find_by_conditions(entities: list, conditions: dict, regex=False, return_slugs=False):
    entities_filtered = []
    for entity in entities:
        if regex:
            match = all(re.search(item[1], entity.get(item[0], '')) is not None for item in conditions.items())
        else:
            match = all(item in entity.items() for item in conditions.items())
        if match:
            entities_filtered.append(entity)

    if return_slugs:
        return [entity.get('slug') for entity in entities_filtered]
    else:
        return [entity.get('id') for entity in entities_filtered]





import logging
import json
from string import Template
import re
import os
from requests import HTTPError

logger = logging.getLogger(__name__)


class RedashEntity:
    __slots__ = 'ent_type', 'id'

    def __init__(self, ent_type, id=None):
        self.ent_type = ent_type
        self.id = id

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.make_uri()}>'
    
    def __str__(self):
        return json.dumps(self.to_dict(), sort_keys=True)

    def __eq__(self, other):
        return all(getattr(self, field, None) == getattr(other, field, None) for field in self.__slots__)

    def make_uri(self):
        return self.ent_type if self.id is None else f'{self.ent_type}/{self.id}'

    def sort_func(self):
        return self.id is None, self.id
    
    @classmethod
    def from_dict(cls, data):
        filtered_dict = {}
        for c in cls.mro():
            if hasattr(c, '__slots__'):
                for field in set(c.__slots__) & set(data):
                    filtered_dict[field] = data.get(field)
        return cls(**filtered_dict)

    @classmethod
    def from_file(cls, path, id):
        entity_dict = {'id': id}
        with open(os.path.join(path, f'{id}.json'), 'r',  encoding='utf-8') as file:
            entity_dict.update(json.load(file))
        return cls.from_dict(entity_dict)

    def to_dict(self, drop_fields=()):
        self_dict = {}
        for field in (set(self.__slots__) - set(drop_fields)):
            temp = getattr(self, field, None)
            if type(temp) == list:
                self_dict[field] = [t.to_dict() if type(t) in (Widget, Visualization, Query) else t for t in temp]
            else:
                self_dict[field] = temp.to_dict() if type(temp) in (Widget, Visualization, Query) else temp
        return self_dict
    
    def to_file(self, path):
        filename = os.path.join(path, f'{self.id}.json')
        with open(filename, 'w', encoding='utf-8') as file:
            json.dump(self.to_dict(), file, sort_keys=True, indent=4)
        return filename
    
    def to_redash(self, redash_session, try_to_update=False):
        uri = self.make_uri() if try_to_update else self.ent_type
        try:
            remote_entity_dict = redash_session.post(uri, self.to_dict())
        except:
            logger.error(f'Не удалось {"обновить" if try_to_update else "создать"} {uri}')
            raise
        return remote_entity_dict


class Taggable(RedashEntity):
    __slots__ = 'tags',

    def __init__(self, ent_type, id=None, tags=None):
        super().__init__(ent_type, id)
        self.tags = set()
        if tags is not None:
            self.add_tags(tags)

    def add_tags(self, tags):
        if type(tags) in (str, int):
            self.tags.add(str(tags))
        elif type(tags) in (set, list, tuple):
            [self.add_tags(tag) for tag in tags]
        else:
            raise TypeError(f'Wrong type for tag {tags}: {type(tags)}')
        return self

    def remove_tags(self, tags):
        if type(tags) in (str, int):
            self.tags.discard(str(tags))
        elif type(tags) in (set, list, tuple):
            [self.remove_tags(tag) for tag in tags]
        else:
            raise TypeError(f'Wrong type for tag {tags}: {type(tags)}')
        return self

    def publish(self, redash_session):
        redash_session.post(self.make_uri(), {'is_draft': False})

    def unpublish(self, redash_session):
        redash_session.post(self.make_uri(), {'is_draft': True})


class Query(Taggable):
    __slots__ = 'data_source_id', 'query', 'name', 'schedule', 'visualizations', 'options'
    
    def __init__(self,
                 data_source_id,
                 query,
                 id=None,
                 name=None,
                 schedule=None,
                 tags=None,
                 visualizations=None,
                 options=None
                 ):
        super().__init__(ent_type='queries', id=id, tags=tags)
        self.data_source_id = data_source_id
        self.query = query
        self.name = name or 'New rtQuery'
        self.schedule = schedule
        self.visualizations = []
        if visualizations is not None:
            for v in visualizations:
                self.add_visualization(Visualization.from_dict(v))
        self.visualizations.sort(key=lambda v: v.sort_func())
        self.options = options or {}

    def match(self, other):
        return self.data_source_id == other.data_source_id and self.query == other.query

    def _update_query_id(self, new_id):
        self.id = new_id
        for v in self.visualizations:
            v.query_id = new_id
        return self
    
    def to_dict(self):
        return super().to_dict()
    
    def to_redash(self, redash_session, try_to_update=False, publish=True):
        remote_query = Query.from_dict(super().to_redash(redash_session, try_to_update))
        if publish:
            remote_query.publish(redash_session)
        self._update_query_id(remote_query.id)
        remote_query_default_vis_id = remote_query.visualizations[0].id
        for v in remote_query.visualizations[1:]:
            try:
                redash_session.delete(v.make_uri())
            except HTTPError:
                print('exception')
        remote_query.visualizations = [remote_query.visualizations[0]]
        visualizations = []
        for v in self.visualizations:
            visualizations.append(Visualization.from_dict(v.to_dict()))  # to avoid mutating of initial vis
        visualizations[0].id = remote_query_default_vis_id
        visualizations[0].query_id = remote_query.id
        remote_query.visualizations[0] = Visualization.from_dict(visualizations[0].to_redash(redash_session,
                                                                                             try_to_update=True))
        for v in visualizations[1:]:
            v.query_id = remote_query.id
            remote_query.visualizations.append(Visualization.from_dict(v.to_redash(redash_session,
                                                                                   try_to_update=False)))
        return remote_query
        
    def to_template(self, param_names):
        return QueryTemplate(self, param_names)
  
    def set_schedule(self, interval_sec):
        self.schedule = {'interval': interval_sec, 
                         'until': None, 
                         'day_of_week': None,
                         'time': None}
        return self
        
    def add_visualization(self, visualization):
        visualization.query_id = self.id
        self.visualizations.append(visualization)
        return visualization

    def replace_sql(self, str_from, str_to, regex=False):
        if regex:
            self.query = re.sub(str_from, str_to, self.query, re.DOTALL)
        else:
            self.query = self.query.replace(str_from, str_to)
        return self
        
        
class Dashboard(Taggable):
    __slots__ = 'slug', 'name', 'widgets', 'queries'
    
    def __init__(self,
                 slug,
                 name=None,
                 id=None,
                 tags=None, 
                 widgets=None):
        super().__init__(ent_type='dashboards', id=id, tags=tags)
        self.slug = slug
        self.name = name or slug
        self.widgets = []
        self.queries = []
        if widgets is not None:
            self.widgets = [Widget.from_dict(w) for w in widgets]
            visualizations = []
            queries = []
            for w in widgets:
                v = w.get('visualization')
                if v is not None:
                    v['query_id'] = v['query']['id']
                    visualizations.append(Visualization.from_dict(v))
                    queries.append(Query.from_dict(v['query']))
            self.queries = list({q.id: q for q in queries}.values())  # deduplication
            for q in self.queries:
                [q.add_visualization(v) for v in visualizations if v.query_id == q.id]
                q.visualizations.sort(key=lambda v: v.sort_func())
            self.queries.sort(key=lambda q: q.sort_func())
            self.widgets.sort(key=lambda w: w.sort_func())


    @classmethod
    def from_file(cls, path, slug):
        entity_dict = {'slug': slug}
        with open(os.path.join(path, f'{slug}.json'), 'r',  encoding='utf-8') as file:
            entity_dict.update(json.load(file))
        return cls.from_dict(entity_dict)

    def _update_id(self, redash_session):
        d = redash_session.get(f'{self.ent_type}/{self.slug}')
        if d['is_archived']:
            logger.error(
                f'Дашборд {redash_session.url}/dashboard/{self.slug} заархивирован и не может быть восстановлен')
            raise UserWarning
        if not d['can_edit']:
            logger.error(f'Нет прав на редактирование дашборда {redash_session.url}/dashboard/{self.slug}')
            raise UserWarning
        self.id = d['id']
        for w in self.widgets:
            w.dashboard_id = self.id
        return self

    def _update_visualization_ids(self, ids_matching):
        for w in self.widgets:
            if w.visualization_id is not None:
                w.visualization_id = ids_matching.get(w.visualization_id)
        return self

    # def to_dict(self):
    #     return super().to_dict()
    
    def to_redash(self, redash_session, try_to_update=False, publish=True):
        if not try_to_update:
            remote_dict = redash_session.post('dashboards', {'name': self.slug})  # try to occupy slug
            remote_dict = redash_session.post(f'dashboards/{remote_dict["id"]}', {'name': self.name})  # rename
            remote_db = Dashboard.from_dict(remote_dict)
        else:
            remote_db = Dashboard.from_dict(super().to_redash(redash_session, try_to_update=True))
        if publish:
            remote_db.publish(redash_session)
        ids_matching = {}
        for q in self.queries:
            remote_q = q.to_redash(redash_session, try_to_update=try_to_update)
            remote_db.queries.append(remote_q)
            for v, remote_v in zip(q.visualizations, remote_q.visualizations):
                ids_matching[v.id] = remote_v.id

        for w in remote_db.widgets:
            try:
                redash_session.delete(w.make_uri())
            except HTTPError:
                print('exception')
        remote_db.widgets = []
        for w in self.widgets:
            w_copy = Widget.from_dict(w.to_dict())  # to avoid mutating of initial widgets
            w_copy.dashboard_id = remote_db.id
            if w_copy.visualization_id is not None:
                w_copy.visualization_id = ids_matching.get(w_copy.visualization_id)
            remote_db.widgets.append(Widget.from_dict(w_copy.to_redash(redash_session,
                                                                       try_to_update=True)))
        return remote_db

    def create_slug(self, redash_session):
        redash_session.post('dashboards', {'name': self.slug})
        
    def to_template(self, param_names):
        return DashboardTemplate(self, param_names)

    def set_schedule(self, interval_sec):
        for q in self.queries:
            q.set_schedule(interval_sec)
        return self

    def change_parameter_level(self, parameter_name, level='dashboard', title=None):
        levels = ('dashboard', 'widget')
        if level not in levels:
            logger.error(f'Допустимые значения level: {levels}')
            return
        for widget in self.widgets:
            if 'parameterMappings' in widget.options.keys():
                if parameter_name in widget.options['parameterMappings'].keys():
                    widget.options['parameterMappings'][parameter_name] = {'type': '{level}-level',
                                                                    'mapTo': parameter_name,
                                                                    'name': parameter_name,
                                                                    'title': title or parameter_name}
        return self

    def publish(self, redash_session):
        redash_session.post(self.make_uri(), {'is_draft': False})

    def unpublish(self, redash_session):
        redash_session.post(self.make_uri(), {'is_draft': True})


class Visualization(RedashEntity):
    __slots__ = 'query_id', 'type', 'options', 'name'
    
    def __init__(self, type, query_id=None, options=None, id=None, name=None):
        super().__init__('visualizations', id)
        self.query_id = query_id
        self.type = type
        self.options = options or {}
        self.name = name or 'New rtVisualization'

    def match(self, other):
        return self.type == other.type and self.options == other.options


class Widget(RedashEntity):
    __slots__ = 'dashboard_id', 'visualization_id', 'text', 'width', 'options'
    
    def __init__(self,
                 dashboard_id,
                 options,
                 visualization_id=None,
                 text=None,
                 id=None,
                 width=None):
        super().__init__('widgets', id)
        self.dashboard_id = dashboard_id
        self.visualization_id = visualization_id
        self.text = text if text != '' else None
        self.options = options
        self.width = width or 1

    @classmethod
    def from_dict(cls, data):
        if 'visualization' in data:
            data['visualization_id'] = data['visualization'].get('id')
        return super().from_dict(data)


class QueryTemplate:
    def __init__(self, query, param_names):
        self.data_source_id = query.data_source_id
        sql = query.query
        for p in param_names:
            pattern = '{{ *' + p + ' *}}'
            sql = re.sub(pattern, f'${p}', sql)
        self.query = Template(sql)
        self.name = query.name
        self.tags = query.tags
        self.visualizations = query.visualizations
        self.options = query.options
        for p in self.options.get('parameters', []):
            if p.get('name') in param_names:
                self.options['parameters'].remove(p)
    
    def render(self, param_dict, custom_tags=None):
        query = Query(data_source_id=self.data_source_id,
                      query=self.query.substitute(param_dict),
                      name=self.name,
                      tags=self.tags,
                      options=self.options)
        query.visualizations = self.visualizations
        if custom_tags is not None:
            query.add_tags(custom_tags)
        return query

    # TODO: add methods for rendering template from dict and saving to dict
    
        
class DashboardTemplate:
    def __init__(self, dashboard, param_names, slug=None):
        self.slug = slug or dashboard.slug
        self.name = dashboard.name
        self.tags = dashboard.tags
        self.widgets = dashboard.widgets
        self.queries = [QueryTemplate(q, param_names) for q in dashboard.queries] 
    
    def render(self, param_dict, slug=None, name=None, custom_tags=None):
        slug = slug or self.slug
        name = name or self.name
        dashboard = Dashboard(slug=slug,
                              name=name,
                              tags=self.tags)
        dashboard.widgets = self.widgets
        dashboard.queries = [q.render(param_dict, custom_tags) for q in self.queries]
        if custom_tags is not None:
            dashboard.add_tags(custom_tags)
        return dashboard



from setuptools import setup

setup(name='redash_tools',
      packages=['redash_tools'],
      version='1.0.2',
      license='MIT',
      description='Tools to backup, batch update, template redash queries and dashboards',
      author='Marina Pavlova',
      author_email='pavlova.marina.v@gmail.com',
      url='http://github.com/pavlova-marina/redash-tools',
      download_url='https://github.com/pavlova-marina/redash-tools/archive/refs/tags/v1.0.2.tar.gz',
      keywords=['redash'],
      install_requires=['requests'],
      python_requires='>=3.6',
      classifiers=[
                  'Development Status :: 3 - Alpha',
                  'Intended Audience :: Developers',
                  'Topic :: Software Development :: Build Tools',
                  'License :: OSI Approved :: MIT License',
                  'Operating System :: OS Independent',
                  'Programming Language :: Python :: 3',
                  'Programming Language :: Python :: 3.4',
                  'Programming Language :: Python :: 3.5',
                  'Programming Language :: Python :: 3.6',
              ]
      )
